using System;
using System.Linq;
using System.Collections;
using System.Collections.Generic;
using System.Reflection;
using UnityEngine;
using UnityEngine.Events;
using UnityEngine.UI;
using TMPro;
using UIEditor;

namespace ObjectEditor
{
    public class ObjectEditorBehaviour : MonoBehaviour
    {
		private GameObject inspected;

		private RectTransform rect;

		private bool hidden;

        private RectTransform container;
		private TextMeshProUGUI hiddenText;
		private ScrollRect scroll;
		private Button goUp;
		private List<GameObject> content = new List<GameObject>();

		private DialogBox activeDialog;

		private ColorBlock defaultColorBlock
		{
			get
			{
				ColorBlock result = new ColorBlock();
				result.colorMultiplier = 1;
				result.fadeDuration = 0.1f;
				result.disabledColor = new Color(0.75f, 0.75f, 0.75f);
				result.highlightedColor = Color.white;
				result.normalColor = Color.white;
				result.pressedColor = new Color(0.75f, 0.75f, 0.75f);
				result.selectedColor = Color.white;
				return result;
			}
		}

		private void Awake()
        {
            rect = GetComponent<RectTransform>();
			inspected = SelectionController.Main.SelectedObjects.First().gameObject;
			CreateEditor();
		}

		private void Update()
        {
			if (goUp && scroll)
				goUp.interactable = scroll.verticalNormalizedPosition != 1f;
		}

        private void CreateEditor()
        {
            transform.position += new Vector3(0, 175f);
            transform.SetParent(UIElements.GetDialogContainer);
            rect.pivot = new Vector2(0.5f, 1f);

            GetComponent<Image>().color = new Color(0.078f, 0.082f, 0.101f);

            #region Hidden text
            hiddenText = UIBuilder.CreateText("", 16, new Color(1, 1, 1, 0.1f), new Vector3(transform.position.x, 185));
            hiddenText.gameObject.AddComponent<TextColorAnimation>().targetColor = new Color(1, 1, 1, 0.1f);
            UnityEngine.Object.Destroy(hiddenText.gameObject.GetComponent<ContentSizeFitter>());
            hiddenText.rectTransform.sizeDelta = new Vector2(260f, 20f);
            hiddenText.enableAutoSizing = true;
            hiddenText.fontSizeMax = 16;
            hiddenText.fontSizeMin = 4;
            hiddenText.maskable = false;
            hiddenText.raycastTarget = false;
            hiddenText.GetComponent<RectTransform>().anchorMax = new Vector2(0.5f, 1f);
            hiddenText.GetComponent<RectTransform>().anchorMin = new Vector2(0.5f, 1f);
            hiddenText.transform.SetParent(transform);
            #endregion

            #region Layout group
            VerticalLayoutGroup verticalGroup = UIBuilder.CreateVerticalLayoutGroup(UIElements.CANVAS_CENTER, Vector2.zero);
            RectTransform groupRect = verticalGroup.GetComponent<RectTransform>();
            groupRect.anchorMin = new Vector2(0, 0);
            groupRect.anchorMax = new Vector2(1, 1);
            groupRect.pivot = new Vector2(0.5f, 1);
            verticalGroup.spacing = 2f;
            RectOffset padding = verticalGroup.padding;
            padding.top = 2;
            padding.bottom = 2;
            verticalGroup.padding = padding;
            verticalGroup.transform.SetParent(transform);
            verticalGroup.gameObject.AddComponent<ContentSizeFitter>().verticalFit = ContentSizeFitter.FitMode.PreferredSize;
            container = groupRect;
			#endregion

			#region Scroll rect
			scroll = gameObject.AddComponent<ScrollRect>();
            scroll.content = groupRect;
            scroll.movementType = ScrollRect.MovementType.Clamped;
            scroll.scrollSensitivity = 35f;
            scroll.horizontal = false;
            scroll.verticalNormalizedPosition = 1f;
			#endregion

			gameObject.AddComponent<Mask>();
			gameObject.AddComponent<StayOnScreen>();
			gameObject.AddComponent<PreventClickDrag>();
			gameObject.AddComponent<DraggablePanel>();
			gameObject.AddComponent<ActionAfterDestroy>().Setup(inspected, () => CloseEditor());
			gameObject.AddComponent<SizeAnimation>();

			#region Editor buttons
			//Close
			CreateEditorButton(ModResources.CloseSprite, () => CloseEditor());
			//Hide
			CreateEditorButton(ModResources.HideSprite, () =>
			{
				hidden = !hidden;
				verticalGroup.gameObject.SetActive(!hidden);
				GetComponent<SizeAnimation>().targetSize = new Vector2(260f, hidden ? 20f : 350f);
				hiddenText.gameObject.GetComponent<TextColorAnimation>().targetColor = new Color(1, 1, 1, hidden ? 0.5f : 0.1f);
				scroll.enabled = !hidden;
				if (!hidden)
					rect.SetAsLastSibling();
			});
			//Go up
			goUp = CreateEditorButton(ModResources.GoUpSprite, () =>
			{
				scroll.verticalNormalizedPosition = 1f;
				goUp.interactable = false;
			});
			#endregion

			UIBuilder.CreateText($"<color=#d1d4e1>Object Editor</color>\n<color=grey><size=12>version {ObjectEditorController.version}\nby Morzzic</size></color>", 16, Color.white, Vector2.zero).transform.SetParent(verticalGroup.transform);

            Refresh();
        }

		int editorButtonsCounter = 0;
		private Button CreateEditorButton(Sprite sprite, UnityAction action)
		{
			editorButtonsCounter++;
			Button EditorButton = UIBuilder.CreateButton(sprite, new Vector3(transform.position.x + 12 - (editorButtonsCounter * 22f), transform.position.y - 10f), new Vector2(20f, 20f));
			EditorButton.colors = defaultColorBlock;
			EditorButton.transform.SetParent(transform);
			EditorButton.GetComponent<RectTransform>().anchorMax = new Vector2(1f, 1f);
			EditorButton.GetComponent<RectTransform>().anchorMin = new Vector2(1f, 1f);
			EditorButton.onClick.AddListener(action);
			return EditorButton;
		}

		public void Refresh() => StartCoroutine(RefreshCoroutine());

		private IEnumerator RefreshCoroutine()
		{
			foreach (GameObject obj in content)
				GameObject.Destroy(obj);
			yield return new WaitForEndOfFrame();
			content.Clear();
			CreateContent();
		}

		private void CreateContent()
        {
			CreateContent(inspected);

			if (inspected.transform.parent)
            {
				if (!inspected.transform.parent.GetComponent<PhysicalBehaviour>())
                {
					CreateContent(inspected.transform.parent.gameObject);
					hiddenText.text = $"{inspected.name} <color=yellow>&</color> {inspected.transform.parent.name}";
				}
			}
			else
				hiddenText.text = inspected.name;
		}

		private void CreateContent(GameObject obj)
		{
			content.Add(UIBuilder.CreateText(obj.name, 14, new Color32(126, 231, 135, 255), container).gameObject);

			content.Add(CreateButton("Find a variable", () =>
			{
				OpenDialog("Enter a component name", "Component name...", "", () =>
				{
					string componentName = activeDialog.EnteredText;
					OpenDialog("Enter a variable name", "Variable name...", "", () =>
					{
						string variableName = activeDialog.EnteredText;
						Debug.Log(componentName + ", " + variableName);
						OpenPropertyDialog(variableName, componentName, null, obj);
					});
				});
			}));
			content.Add(CreateButton("Add component", () =>
			{
				OpenDialog("Enter a component name", "Component name...", "", () =>
				{
					string componentName = activeDialog.EnteredText;
					Type type = Type.GetType("UnityEngine." + componentName + ", UnityEngine");
					if (type == null)
					{
						type = Type.GetType(componentName + ", UnityEngine");
						if (type == null)
                        {
							type = Type.GetType(componentName + ", Assembly-CSharp");
							if (type == null)
								type = Type.GetType(componentName);
						}
					}

					if (type == null)
						ModAPI.Notify($"<color=red>Component <b>{componentName}</b> not found!</color>");
					else
					{
						obj.AddComponent(type);
						Refresh();
						ModAPI.Notify($"<color=green>The <b>{componentName}</b> component has been added.</color>");
					}
				});
			}));

			var flags = BindingFlags.Instance | BindingFlags.Public;

			foreach (Component component in obj.GetComponents<Component>())
			{
				List<GameObject> buttons = new List<GameObject>();
				Transform text = UIBuilder.CreateText(component.GetType().Name, 10, new Color(0.345f, 0.360f, 0.435f), container).transform;
				content.Add(text.gameObject);

				foreach (PropertyInfo property in component.GetType().GetProperties(flags))
				{
					if (ObjectEditorController.allowedTypes.Contains(property.PropertyType))
					{
						GameObject button = CreateButton("Set " + Utilities.FormatToSentence(property.Name), () => OpenPropertyDialog(property.Name, component.GetType().Name, property.GetValue(component), obj));
						button.AddComponent<SizeAnimation>();
						buttons.Add(button);
					}
				}

				foreach (FieldInfo field in component.GetType().GetFields(flags))
				{
					if (ObjectEditorController.allowedTypes.Contains(field.FieldType))
					{
						GameObject button = CreateButton("Set " + Utilities.FormatToSentence(field.Name), () => OpenPropertyDialog(field.Name, component.GetType().Name, field.GetValue(component), obj));
						button.AddComponent<SizeAnimation>();
						buttons.Add(button);
					}
				}

				content.AddRange(buttons);

				#region Switch button
				Button SwitchButton = UIBuilder.CreateButton(new Vector3(text.position.x + 118, text.position.y), new Vector2(20f, 20f));
				SwitchButton.colors = defaultColorBlock;
				SwitchButton.image.color = new Color(0.290f, 0.294f, 0.333f);
				SwitchButton.transform.SetParent(text);
				SwitchGameObjects sgo = SwitchButton.gameObject.AddComponent<SwitchGameObjects>();
				sgo.objects = buttons;
				Image img = UIBuilder.CreateImage(ModResources.SwitchSprite, SwitchButton.transform.position, new Vector2(20f, 20f));
				img.transform.SetParent(SwitchButton.transform);
				img.raycastTarget = false;
				sgo.image = img;
				#endregion

				#region Remove button
				Button RemoveButton = UIBuilder.CreateButton(ModResources.RemoveSprite, new Vector3(text.position.x + 96, text.position.y), new Vector2(20f, 20f));
				RemoveButton.colors = defaultColorBlock;
				RemoveButton.transform.SetParent(text);
				RemoveButton.onClick.AddListener(() =>
				{
					OpenDialog($"Are you sure you want to remove the <b>{component.GetType().Name}</b> component?", "", "", () =>
					{
						UnityEngine.Object.Destroy(component);
						Refresh();
					});
					activeDialog.ShowTextBox = false;
				});
				#endregion
			}
		}

		private void OpenDialog(string title, string placeholder, string enteredText, UnityAction action)
		{
			DialogBox dialog = (DialogBox)null;
			dialog = DialogBoxManager.TextEntry(title, placeholder, new[]
			{
				new DialogButton("OK", true, action),
				new DialogButton("Cancel", true, () => dialog.Close())
			});
			dialog.EnteredText = enteredText;
			activeDialog = dialog;
		}

		private void OpenPropertyDialog(string variableName, string componentName, object value, GameObject obj)
		{
			Component component = obj.GetComponent(componentName);

			if (!component)
			{
				ModAPI.Notify("<color=red>Unknown component.</color>");
				return;
			}

			var flags = BindingFlags.Instance | BindingFlags.Public;

			UnityAction<string> action;

			PropertyInfo property = component.GetType().GetProperty(variableName, flags);
			FieldInfo field = component.GetType().GetField(variableName, flags);

			object result;
			string errorMessage = "<color=red>An error has occurred. Check out the help with entering values on the mod page in the Steam Workshop.</color>";

			if (property != null)
			{
				action = (str) =>
				{
					if (Utilities.TryConvertToObject(str, property.PropertyType, out result))
					{
						property.SetValue(component, result);
						ModAPI.Notify($"<color=green>Variable <b>{componentName}.{variableName}</b> is set to <b>{Utilities.ConvertToString(property.GetValue(component))}</b>.</color>");
					}
					else
						ModAPI.Notify(errorMessage);
				};
			}
			else if (field != null)
			{
				action = (str) =>
				{
					if (Utilities.TryConvertToObject(str, field.FieldType, out result))
					{
						field.SetValue(component, result);
						ModAPI.Notify($"<color=green>Variable <b>{componentName}.{variableName}</b> is set to <b>{Utilities.ConvertToString(field.GetValue(component))}</b>.</color>");
					}
					else
						ModAPI.Notify(errorMessage);
				};
			}
			else
				action = (str) => ModAPI.Notify($"<color=red>Variable \"{componentName}.{variableName}\' not found.</color>");

			string type = value.GetType().Name;
			type = type.Replace("String", "string");
			type = type.Replace("Boolean", "bool");
			type = type.Replace("Int32", "int");
			type = type.Replace("Single", "float");

			OpenDialog($"Edit {Utilities.FormatToSentence(variableName)}" + (value == null ? string.Empty : $"\n<size=24>({type})</size>"), $"Enter {type}", Utilities.ConvertToString(value), () => action?.Invoke(activeDialog.EnteredText));
		}

		private GameObject CreateButton(string text, UnityAction action)
		{
			Button button = UIBuilder.CreateButton(Vector3.zero, new Vector2(256f, 30f));
			button.image.color = new Color(0.290f, 0.294f, 0.333f);
			button.colors = defaultColorBlock;
			button.transform.SetParent(container);
			TextMeshProUGUI tmp = UIBuilder.CreateText(text, 12, new Color32(209, 212, 225, 255), button.transform.position);
			tmp.transform.SetParent(button.transform);
			button.gameObject.AddComponent<Mask>();
			button.onClick.AddListener(action);
			return button.gameObject;
		}

		public void CloseEditor()
        {
			ObjectEditorController.inspectedObjects.Remove(inspected);
            GameObject.Destroy(gameObject);
        }
    }
}