using System;
using System.Linq;
using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.Events;

namespace ObjectEditor
{
    public static class Utilities
    {
		public static string FormatToSentence(string str)
        {
			return new string(str.SelectMany((c, i) => i > 0 && char.IsUpper(c) ? new[] { ' ', c } : new[] { c }).ToArray()).ToLower();
		}

		public static string ConvertToString(object value)
		{
			if (value == null)
            {
				return "";
            }
			else if (value.GetType() == typeof(int))
			{
				return (value.ToString());
			}
			else if (value.GetType() == typeof(float))
			{
				return (value.ToString().Replace(',', '.'));
			}
			else if (value.GetType() == typeof(Color))
			{
				Color color = (Color)value;
				return $"{(color.r * 255).ToString().Replace(',', '.')}, {(color.g * 255).ToString().Replace(',', '.')}, {(color.b * 255).ToString().Replace(',', '.')}" + (color.a == 1 ? string.Empty : $", {(color.a * 255).ToString().Replace(',', '.')}");
			}
			else if (value.GetType() == typeof(Vector3))
			{
				Vector3 vector = (Vector3)value;
				return $"{(vector.x).ToString().Replace(',', '.')}, {(vector.y).ToString().Replace(',', '.')}, {(vector.z).ToString().Replace(',', '.')}";
			}
			else if (value.GetType() == typeof(Vector2))
			{
				Vector2 vector = (Vector2)value;
				return $"{(vector.x).ToString().Replace(',', '.')}, {(vector.y).ToString().Replace(',', '.')}";
			}
			else if (value.GetType() == typeof(string))
			{
				return value.ToString();
			}
			else if (value.GetType() == typeof(bool))
			{
				return (bool)value ? "True" : "False";
			}
			else if (value.GetType() == typeof(Cartridge))
			{
				return $"{((Cartridge)value).Damage}; {((Cartridge)value).Recoil}; {((Cartridge)value).ImpactForce}; {((Cartridge)value).StartSpeed}; {((Cartridge)value).PenetrationRandomAngleMultiplier}";
			}
			else
			{
				return "";
			}
		}

		public static bool TryConvertToObject(string str, Type type, out object value)
		{
			if (type == typeof(int))
			{
				try
                {
					value = ToInt(str);
				}
				catch
                {
					value = null;
				}
			}
			else if (type == typeof(float))
			{
				try
				{
					value = ToFloat(str);
				}
				catch
				{
					value = null;
				}
			}
			else if (type == typeof(Color))
			{
				try
				{
					value = ToColor(str);
				}
				catch
				{
					value = null;
				}
			}
			else if (type == typeof(Vector3))
			{
				try
				{
					value = ToVector3(str);
				}
				catch
				{
					value = null;
				}
			}
			else if (type == typeof(Vector2))
			{
				try
				{
					value = ToVector2(str);
				}
				catch
				{
					value = null;
				}
			}
			else if (type == typeof(string))
			{
				try
				{
					value = str;
				}
				catch
				{
					value = null;
				}
			}
			else if (type == typeof(bool))
			{
				try
				{
					value = ToBool(str);
				}
				catch
				{
					value = null;
				}
			}
			else if (type == typeof(Cartridge))
			{
				try
				{
					value = ToCartridge(str);
				}
				catch
				{
					value = null;
				}
			}
			else
				value = null;

			return value != null;
		}

		private static bool ToBool(string str)
		{
			return str.ToLower().Contains("true");
		}

		private static int ToInt(string str)
		{
			return int.Parse(str);
		}

		private static float ToFloat(string str)
		{
			return Convert.ToSingle(str.Replace('.', ','));
		}

		private static Color ToColor(string str)
		{
			string[] split = str.Split(", "[0]);
			Color result = new Color();
			result.r = ToFloat(split[0]) / 255;
			result.g = ToFloat(split[1]) / 255;
			result.b = ToFloat(split[2]) / 255;
			result.a = (split.Length > 3) ? (ToFloat(split[3]) / 255) : 1;
			return result;
		}

		private static Vector3 ToVector3(string str)
		{
			string[] split = str.Split(", "[0]);
			Vector3 result = new Vector3();
			result.x = ToFloat(split[0]);
			result.y = ToFloat(split[1]);
			result.z = (split.Length > 2) ? (ToFloat(split[2])) : 0;
			return result;
		}

		private static Vector2 ToVector2(string str)
		{
			string[] split = str.Split(", "[0]);
			Vector2 result = new Vector2();
			result.x = ToFloat(split[0]);
			result.y = ToFloat(split[1]);
			return result;
		}

		private static Cartridge ToCartridge(string str)
        {
			string[] split = str.Split("; "[0]);
			Cartridge result = new Cartridge();
			result.Damage = ToFloat(split[0]);
			result.Recoil = ToFloat(split[1]);
			result.ImpactForce = ToFloat(split[2]);
			result.StartSpeed = ToFloat(split[3]);
			result.PenetrationRandomAngleMultiplier = ToFloat(split[4]);
			return result;
		}
	}
}
